/**
                              Iberia MUD Engine
    Copyright (C) 2009, 2010, 2011 Joo Afonso <joaodiasafonso@gmail.com>

    This file is part of Iberia MUD Engine.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    Iberia MUD Engine is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Iberia MUD Engine.  If not, see <http://www.gnu.org/licenses/>.
 */
package ime.utils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Scanner;
import java.util.Timer;
import java.util.TimerTask;

import ime.cmds.admin.Translation;
import ime.interfaces.ILivingSet;
import ime.main.Main;
import ime.utils.Utils.TranslationLanguage;

/**
 * @author Joo Afonso <joaodiasafonso@gmail.com>, CAP Inf
 */
public class IMETranslator {
	static final int MAX_HUNGUP_TIME=5 * 1000;//in seconds
	static final String ERROR="Iberia MUD Engine has found an error in translation system and has disabled it.";
	Process leccomm;
	Scanner stdIn;
	PrintWriter stdOut;
	String LECPath;
	boolean isTranslating;
	TranslationLanguage stdOr;
	TranslationLanguage stdDest;
	int cacheSize;
	Map<String,LinkedHashMap<String,String>> cache;
	private static final float loadFactor=0.75F;

	/**
	 * This object creates a leccomm.exe process after a path to LEC LogoTrans.exe
	 * file is given.
	 * This exe is a small app that opens a LEC LogoTrans.exe
	 * app, from LEC Power Translator 15, and communicates with it in order to send
	 * text and obtain it's translation by just using stdIn and stdOut channels
	 * @param cacheSize Size of cache, in number of entries, per pair of 
	 * languages; 0 if no cache wanted
	 */
	public IMETranslator(int cacheSize) {
		leccomm=null;
		stdIn=null;
		stdOut=null;
		LECPath=null;
		isTranslating=false;
		stdOr=null;
		stdDest=null;
		this.cacheSize=cacheSize;
		cache=null;
	}
	
	/**
	 * Sets up cache. Translator uses it in order to avoid to repeat translations if
	 * already did it once (considering also cache size) 
	 */
	public void setupCache() {
		cache=new HashMap<String,LinkedHashMap<String,String>>();
		for(TranslationLanguage a:TranslationLanguage.values()) {
			if(a!=TranslationLanguage.ENGLISH) {
				LinkedHashMap<String,String> lhm1=
					new LinkedHashMap<String,String>((int)((float)cacheSize/loadFactor), loadFactor, true) {
						private static final long serialVersionUID = 1L;

						protected boolean removeEldestEntry(Map.Entry<String,String> eldest) {
							return size() > cacheSize;
						}
				};
				LinkedHashMap<String,String> lhm2=
					new LinkedHashMap<String,String>((int)((float)cacheSize/loadFactor), loadFactor, true) {
					private static final long serialVersionUID = 1L;

					protected boolean removeEldestEntry(Map.Entry<String,String> eldest) {
						return size() > cacheSize;
					}
				};
				cache.put(a.toString()+"to"+TranslationLanguage.ENGLISH, lhm1);
				cache.put(TranslationLanguage.ENGLISH+"to"+a.toString(), lhm2);
			}
		}
	}
	
	/**
	 * Closes translator system and service. Does it by setting turning off translator
	 * system flag and destroying the leccomm process 
	 */
	public void close() {
		this.setTranslating(false);
		if(this.isServiceRunning())
			leccomm.destroy();
		stdIn=null;
		stdOut=null;
		leccomm=null;
	}
	
	/**
	 * Start translation apps: leccomm communicator, and LEC Power Translator. 
	 * leccomm receives input from MUD and passes to Translator, and few time later
	 * captures Translator output
	 * @param LECPath Path to LEC Power Translator: LogoTrans.exe
	 */
	public void start(String LECPath) {
		if(LECPath!=null) {
			try {
				this.LECPath=LECPath;
				leccomm = Runtime.getRuntime().exec("leccomm "+this.LECPath);
				stdIn=new Scanner(new BufferedReader(new InputStreamReader(leccomm.getInputStream())));
				stdIn.useDelimiter("\n");
				stdOut= new PrintWriter(leccomm.getOutputStream());
				stdOr=TranslationLanguage.ENGLISH;
				stdDest=TranslationLanguage.PORTUGUESE;
				if(this.isServiceRunning()) {
					this.setTranslating(true);
					setupCache();
				}
			}
			catch (IOException e) {
				System.out.println("Error opening LECComm.exe.");
				e.printStackTrace();
			}
		}
	}

	/**
	 * Translate text from ENGLISH language to dest language
	 * @param text Text to translate
	 * @param or From language
	 * @param dest To Language
	 * @param ind Living doing de translation
	 * @param scriptExecution TRUE, if is script execution
	 * @return Translated text
	 */
	public String translate(String text, TranslationLanguage dest, ILivingSet ind, boolean scriptExecution) {
		return this.translate(text, TranslationLanguage.ENGLISH, dest, ind, scriptExecution);
	}

	/**
	 * Translate text from or language to dest language
	 * @param text Text to translate
	 * @param or From language
	 * @param dest To Language
	 * @param ind Living doing de translation
	 * @param scriptExecution TRUE, if is script execution
	 * @return Translated text
	 */
	public synchronized String translate(String text, TranslationLanguage or, 
			TranslationLanguage dest, final ILivingSet ind, final boolean scriptExecution) {
		String tempText=text;
		String result=null;
		if(this.isServiceRunning()) {
			//check cache first
			if(cache!=null)
				result=cache.get(or.toString()+"to"+dest.toString()).get(text);
			if(result==null) {
				if(stdOr!=or || stdDest!=dest) {
					if(or==stdDest && dest==stdOr) {
						stdOut.println("SWAP_LANG");
						stdOut.flush();
					}
					else {
						if(stdOr!=or && stdDest!=dest) {
							if(or==stdDest) {
								stdOut.println("TRANSLATE_TO "+dest.toString());
								stdOut.flush();
								stdOut.println("TRANSLATE_FROM "+or.toString());
								stdOut.flush();
							}
							else {
								stdOut.println("TRANSLATE_FROM "+or.toString());
								stdOut.flush();
								stdOut.println("TRANSLATE_TO "+dest.toString());
								stdOut.flush();
							}
						}
						else {
							if(stdOr!=or) {
								stdOut.println("TRANSLATE_FROM "+or.toString());
								stdOut.flush();
							}
							else {
								stdOut.println("TRANSLATE_TO "+dest.toString());
								stdOut.flush();
							}
						}
					}
					stdOr=or;
					stdDest=dest;
				}

				//send text to translate
				stdOut.println(tempText);
				stdOut.flush();				
				try {
					Timer x=new Timer();
					x.schedule(new TimerTask() {
					    @Override
					    public void run() {
					    	if(isServiceRunning())
					    		imeTranslatorAbort(ind, scriptExecution);
					    }
					}, MAX_HUNGUP_TIME);
					result=stdIn.next();
					//result comes from leccomm with something at the end, so erase it
					//without this, a line break would be added at the end 
					if(result.length()>0)
						result=result.substring(0, result.length()-1);
					x.cancel();
				} catch (Exception e) {
					if(isServiceRunning())
						imeTranslatorAbort(ind, scriptExecution);
					return text;
				}				
				//add this recent translation to cache
				if(cache!=null)
					cache.get(or.toString()+"to"+dest.toString()).put(text, result);
			}
		}
		else {
			ind.setTranslationLanguage(TranslationLanguage.ENGLISH);
			result=text;
		}
		return result;
	}
	
	/**
	 * Method to be executed when an error arises. At the moment, those errors are mostly
	 * due to big delays between app communications
	 * @param ind Living wanting a translation
	 * @param scriptExecution TRUE, if is script execution
	 */
	private void imeTranslatorAbort(ILivingSet ind, boolean scriptExecution) {
		this.close();
		Translation t=((Translation)ind.getCommand(Main.dm.getCs().getCommands(),"translation"));
		t.updateTranslationSystemOffInPlayers();
		ind.getCm().sendInfo(ERROR, false, false, false, false);
		t.infoIberiaAboutUpdate(false,ind,Translation.TRANSLATION_SYSTEM_OFF, scriptExecution);
	}

	/**
	 * Check if translation system is activated/turned on
	 * @return TRUE, if is translating
	 */
	public boolean isTranslating() {
		return isTranslating;
	}

	/**
	 * Update if translation system is turned on or off.
	 * @param isTranslating TRUE, if is turned on
	 */
	public void setTranslating(boolean isTranslating) {
		this.isTranslating = isTranslating;
	}
	
	/**
	 * Check if translator service is running on server. Does this by checking if leccomm
	 * process is in memory running 
	 * @return TRUE, if is running
	 */
	public boolean isServiceRunning() {
		return leccomm!=null;
	}
}